# -*- coding: utf-8 -*-

import os
import hashlib

from typing import Optional, List, Tuple

from PyQt5 import uic
from PyQt5.QtCore import QRegularExpression, QCoreApplication
from PyQt5.QtGui import QRegularExpressionValidator
from PyQt5.QtWidgets import QDialog, QWidget, QLineEdit, QRadioButton, QLabel, QDialogButtonBox, QComboBox, QMessageBox

from fly_csp_plugins.signature_mark_widget.template_engine import TemplateData

translate = QCoreApplication.translate

dialog_ui = os.path.join(os.path.dirname(__file__), 'ui', 'createtemplatedialog.ui')
UiFormClass, UiFormBase = uic.loadUiType(dialog_ui)  # type: QDialog, QDialog


class CreateTemplateDialog(UiFormClass, UiFormBase):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent=parent)
        self.templates_combobox = None  # type: Optional[QComboBox]
        self.name_label = None  # type: Optional[QLabel]
        self.name_line_edit = None  # type: Optional[QLineEdit]
        self.global_radio_button = None  # type: Optional[QRadioButton]
        self.user_radio_button = None  # type: Optional[QRadioButton]
        self.button_box = None  # type: Optional[QDialogButtonBox]
        self.setupUi(self)
        self.__setup()

    @property
    def name(self) -> str:
        return self.name_line_edit.text()

    @property
    def dir_name(self) -> str:
        return hashlib.md5(self.name.encode()).hexdigest()

    @property
    def selected_template(self) -> Optional[TemplateData]:
        return self.templates_combobox.currentData()

    @selected_template.setter
    def selected_template(self, template: TemplateData) -> None:
        self.templates_combobox.setCurrentIndex(self.templates_combobox.findData(template))

    @property
    def templates(self) -> List[Tuple]:
        return [(self.templates_combobox.itemText(index), self.templates_combobox.itemData(index))
                for index in range(self.templates_combobox.count())]

    @templates.setter
    def templates(self, templates: List[Tuple]) -> None:
        self.templates_combobox.clear()
        for text, template in templates:
            self.templates_combobox.addItem(text, template)

    @property
    def selected_location(self) -> TemplateData.Location:
        if self.global_radio_button.isChecked():
            return TemplateData.Location.Global
        elif self.user_radio_button.isChecked():
            return TemplateData.Location.User
        else:
            return TemplateData.Location.Unknown

    def accept(self) -> None:
        if TemplateData(name=self.dir_name, location=self.selected_location).exist:
            QMessageBox.critical(self,
                                 translate('text', 'Template exists'),
                                 translate('text', 'Template with current dir name already exists'))
            return

        super().accept()

    def __setup(self) -> None:
        validator = QRegularExpressionValidator(QRegularExpression(r'(?:(?![/]).)+'), parent=self)
        self.name_line_edit.setValidator(validator)

        self.name_line_edit.textChanged.connect(self.__on_name_line_edit_text_changed)

        self.__check_data()

    def __on_name_line_edit_text_changed(self) -> None:
        self.__check_data()

    def __check_data(self) -> None:
        valid = True

        if not len(self.name_line_edit.text()):
            self.name_line_edit.setPlaceholderText(translate('text', 'empty name'))
            valid = False
        else:
            self.name_line_edit.setPlaceholderText(None)

        self.button_box.button(QDialogButtonBox.Ok).setEnabled(valid)
