# -*- coding: utf-8 -*-

from typing import List, Optional

from PyQt5.QtCore import Qt, QSize, pyqtSignal
from PyQt5.QtGui import QImage, QPainter, QPixmap
from PyQt5.QtWidgets import QLabel, QWidget

from libastra_csp.certificate import Certificate
from fly_csp_plugins.signature_mark_widget.template_engine import TemplateEngine
from fly_csp_plugins.signature_mark_widget.template_engine import CertificateKeyNotFoundError


class MarkLabel(QLabel):
    content_changed = pyqtSignal()

    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.__template_engine = TemplateEngine()
        self.__template_label = QLabel()
        self.__scale_factor = 100

    def __setup(self) -> None:
        self.__template_label.setTextFormat(Qt.MarkdownText)

    @property
    def template_text(self) -> Optional[str]:
        return self.__template_engine.template

    @template_text.setter
    def template_text(self, text: str) -> None:
        self.__template_engine.template = text

    @property
    def certificate(self) -> Optional[Certificate]:
        return self.__template_engine.certificate

    @certificate.setter
    def certificate(self, certificate: Certificate) -> None:
        self.__template_engine.certificate = certificate

    @property
    def style_sheet(self) -> str:
        return self.__template_label.styleSheet()

    @style_sheet.setter
    def style_sheet(self, text: str) -> None:
        self.__template_label.setStyleSheet(text)

    @property
    def scale_factor(self) -> int:
        return self.__scale_factor

    @scale_factor.setter
    def scale_factor(self, value: int) -> None:
        self.__scale_factor = value

    @property
    def errors(self) -> List[CertificateKeyNotFoundError]:
        return self.__template_engine.errors

    def update_content(self) -> None:
        self.__template_label.setText(self.__template_engine.render())
        self.__template_label.resize(self.__template_label.sizeHint())
        self.__scale(self.__scale_factor)
        self.content_changed.emit()

    def save_as_image(self, filepath: str) -> bool:
        return self.__template_label.grab().toImage().save(filepath)

    def __scale(self, percentage: int) -> None:
        factor = percentage / 100.0
        width = int(self.__percentage_value(percentage, self.__template_label.width()))
        height = int(self.__percentage_value(percentage, self.__template_label.height()))
        self.setPixmap(self.__render_pixmap(factor, width, height))
        self.resize(width, height)

    # TODO: Возможно, эта реализация подойдёт лучше при <60%
    def __scale_simple(self, percentage: int) -> None:
        width = int(self.__percentage_value(percentage, self.__template_label.width()))
        height = int(self.__percentage_value(percentage, self.__template_label.height()))
        pixmap = self.__template_label.grab()
        scaled_pixmap = pixmap.scaled(width, height, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        self.setPixmap(scaled_pixmap)
        self.resize(width, height)

    def __percentage_value(self, percent: int, whole: int) -> float:
        return (percent * whole) / 100.0

    def __render_pixmap(self, factor, width, height) -> QPixmap:
        bitmap = QImage(QSize(width, height), QImage.Format_ARGB32_Premultiplied)
        painter = QPainter(bitmap)
        painter.scale(factor, factor)
        painter.setCompositionMode(QPainter.CompositionMode_Source)
        self.__template_label.render(painter)
        painter.end()
        return QPixmap.fromImage(bitmap)
