# -*- coding: utf-8 -*-

import os

from typing import Optional, Tuple

from popplerqt5 import Poppler
from reportlab.pdfgen.canvas import Canvas
from reportlab.lib.utils import ImageReader
from PyPDF2 import PdfFileReader

from PyQt5 import uic
from PyQt5.QtCore import QFileInfo, QProcess, QSize, QCoreApplication, QLocale
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget, QToolBar, QAction, QWidgetAction, QPushButton
from PyQt5.QtWidgets import QMessageBox, QComboBox
from PyQt5.QtGui import QPixmap, QScreen

from fly_csp_plugins.signature_mark_widget.pages_spinbox import PagesSpinBox
from fly_csp_plugins.signature_mark_widget.zoom_combobox import ZoomComboBox
from fly_csp_plugins.signature_mark_widget.scroll_area import ScrollArea
from fly_csp_plugins.signature_mark_widget.view_widget import ViewLabel
from fly_csp_plugins.signature_mark_widget.mark_widget import MarkWidget
from fly_csp_plugins.signature_mark_widget.watermark_creation_thread import WatermarkCreationThread
from fly_csp_plugins.signature_mark_widget.wait_thread_dialog import WaitThreadDialog
from fly_csp_plugins.signature_mark_widget.template_engine import TemplateData
from fly_csp_plugins.signature_mark_widget.templates_settings_dialog import TemplatesSettingsDialog
from fly_csp_plugins.signature_mark_widget.signature_mark_settings import SignatureMarkUserSettings
from libfly_csp.base_main_window import BaseMainWindow
from libastra_csp.process import CompletedResult
from libastra_csp.certificate import Certificate
from libfly_csp.pdf_conversion_process import PdfConversionProcess
from libfly_csp.interactive_process_dialog import InteractiveProcessDialog
from libastra_csp.utils import Utils
from libastra_csp.container_creation_thread import ContainerCreationThread

translate = QCoreApplication.translate

dialog_ui = os.path.join(os.path.dirname(__file__), 'ui', 'mainwidget.ui')
UiFormClass, UiFormBase = uic.loadUiType(dialog_ui)  # type: QMainWindow, QMainWindow


class Document:
    def __init__(self, document: Poppler.Document, path: str):
        self.p_document = document
        self.path = path


class SignatureMarkWidget(UiFormClass, UiFormBase):
    def __init__(self, main_window: BaseMainWindow, parent: QWidget = None) -> None:
        super().__init__(parent=parent)
        self.scroll_area = None  # type: Optional[ScrollArea]
        self.view_label = None  # type: Optional[ViewLabel]
        self.navigation_toolbar = None  # type: Optional[QToolBar]
        self.actions_toolbar = None  # type: Optional[QToolBar]
        self.first_page_action = None  # type: Optional[QAction]
        self.previous_page_action = None  # type: Optional[QAction]
        self.next_page_action = None  # type: Optional[QAction]
        self.last_page_action = None  # type: Optional[QAction]
        self.zoom_out_action = None  # type: Optional[QAction]
        self.zoom_in_action = None  # type: Optional[QAction]
        self.templates_settings_action = None  # type: Optional[QAction]
        self.setupUi(self)
        self.current_page_spinbox = PagesSpinBox()
        self.current_zoom_combobox = ZoomComboBox()
        self.template_combobox = QComboBox()
        self.apply_stamp_push_button = QPushButton(translate('text', 'Apply'))
        self.__main_window = main_window
        self.__document = None  # type: Optional[Document]
        self.__mark_widget = MarkWidget(parent=self.view_label)
        self.__pdf_conversion_process = PdfConversionProcess()
        self.__settings = SignatureMarkUserSettings()
        self.__setup()

    @property
    def selected_template(self) -> Optional[TemplateData]:
        return self.template_combobox.currentData()

    def __setup(self) -> None:
        current_page_action = QWidgetAction(self.navigation_toolbar)
        current_page_action.setDefaultWidget(self.current_page_spinbox)
        self.navigation_toolbar.insertAction(self.next_page_action, current_page_action)

        current_zoom_action = QWidgetAction(self.navigation_toolbar)
        current_zoom_action.setDefaultWidget(self.current_zoom_combobox)
        self.navigation_toolbar.insertAction(self.zoom_in_action, current_zoom_action)

        self.first_page_action.triggered.connect(self.__on_first_page_action_triggered)
        self.previous_page_action.triggered.connect(self.__on_previous_page_action_triggered)
        self.next_page_action.triggered.connect(self.__on_next_page_action_triggered)
        self.last_page_action.triggered.connect(self.__on_last_page_action_triggered)
        self.zoom_out_action.triggered.connect(self.__on_zoom_out_action_triggered)
        self.zoom_in_action.triggered.connect(self.__on_zoom_in_action_triggered)
        self.scroll_area.wheel_up_triggered.connect(self.__scroll_area_wheel_up_triggered)
        self.scroll_area.wheel_down_triggered.connect(self.__scroll_area_wheel_down_triggered)
        self.templates_settings_action.triggered.connect(self.__on_templates_settings_action_triggered)
        self.template_combobox.currentIndexChanged.connect(self.__on_template_combobox_current_index_changed)
        self.apply_stamp_push_button.clicked.connect(self.__apply_stamp_push_button_clicked)
        self.current_page_spinbox.editingFinished.connect(self.__on_current_page_spinbox_editing_finished)
        self.current_zoom_combobox.currentIndexChanged.connect(self.__current_zoom_combobox_current_index_changed)
        self.view_label.has_resize.connect(self.__on_view_label_has_resize)
        self.__main_window.tabWidget.currentChanged.connect(self.__main_window_tab_widget_current_changed)
        self.__pdf_conversion_process.completed.connect(self.__on_pdf_conversion_process_completed)
        self.__main_window.selected_document_changed.connect(self.__on_main_window_selected_document_changed)
        self.__main_window.comboBoxCertificates.currentIndexChanged.connect(
            self.__combobox_certificates_current_index_changed)

        self.__main_window.settings_menu.addAction(self.templates_settings_action.text(),
                                                   self.__menu_templates_settings_action_triggered)

        template_action = self.actions_toolbar.insertWidget(self.templates_settings_action, self.template_combobox)
        self.actions_toolbar.insertWidget(template_action, self.apply_stamp_push_button)

        self.__block_control_widgets()

        self.__update_templates()

    def __menu_templates_settings_action_triggered(self) -> None:
        dialog = TemplatesSettingsDialog(certificates=self.__main_window.certificates,
                                         selected_certificate=self.__main_window.selected_certificate,
                                         main_window=self.__main_window)
        dialog.set_selected_template(self.selected_template)
        dialog.exec()
        self.__update_templates()

        if self.__main_window.tabWidget.currentWidget() != self:
            self.__unload_document()
        else:
            self.__update_templates()

    def __on_first_page_action_triggered(self, _: bool) -> None:
        self.__load_page(1)

    def __on_previous_page_action_triggered(self, _: bool) -> None:
        self.__load_page(self.current_page_spinbox.value() - 1)

    def __on_next_page_action_triggered(self, _: bool) -> None:
        self.__load_page(self.current_page_spinbox.value() + 1)

    def __on_last_page_action_triggered(self, _: bool) -> None:
        if self.__document is None:
            return

        self.__load_page(self.__document.p_document.numPages())

    def __on_zoom_out_action_triggered(self, _: bool) -> None:
        self.__set_zoom_index(self.current_zoom_combobox.currentIndex() - 1)

    def __on_zoom_in_action_triggered(self, _: bool) -> None:
        self.__set_zoom_index(self.current_zoom_combobox.currentIndex() + 1)

    def __scroll_area_wheel_up_triggered(self) -> None:
        self.__set_zoom_index(self.current_zoom_combobox.currentIndex() + 1)

    def __scroll_area_wheel_down_triggered(self) -> None:
        self.__set_zoom_index(self.current_zoom_combobox.currentIndex() - 1)

    def __on_templates_settings_action_triggered(self) -> None:
        dialog = TemplatesSettingsDialog(certificates=self.__main_window.certificates,
                                         selected_certificate=self.__main_window.selected_certificate,
                                         main_window=self.__main_window)
        dialog.set_selected_template(self.selected_template)
        dialog.exec()
        self.__update_templates()

    def __on_template_combobox_current_index_changed(self, _: int) -> None:
        if not self.selected_template:
            return

        if self.selected_template.exist:
            self.__settings.last_template = self.selected_template.path

        self.__load_mark(certificate=self.__main_window.selected_certificate,
                         template=self.selected_template)

    def __update_templates(self) -> None:
        self.template_combobox.blockSignals(True)
        selected_template_path = self.selected_template.path if self.selected_template else None
        self.template_combobox.clear()

        for template in TemplateData.get_all_templates():
            name = template.read_name(Utils.get_lang())
            self.template_combobox.addItem(name if name else template.dir_name, template)

        if selected_template_path:
            indexes = [index for index in range(self.template_combobox.count())
                       if self.template_combobox.itemData(index).path == selected_template_path]
        else:
            indexes = [index for index in range(self.template_combobox.count())
                       if self.template_combobox.itemData(index).path == self.__settings.last_template]
        if indexes and indexes[0] >= 0:
            self.template_combobox.setCurrentIndex(indexes[0])

        if self.selected_template and self.selected_template.exist:
            self.__settings.last_template = self.selected_template.path

        self.template_combobox.blockSignals(False)
        self.__load_mark(certificate=self.__main_window.selected_certificate,
                         template=self.selected_template)

    def __create_preview_file(self) -> None:
        if self.__document is None:
            return

        self.__main_window.cache.create_dir()

        watermark_png_path = self.__main_window.cache.cache_path() + '/watermark.png'
        if not self.__mark_widget.mark_label.save_as_image(watermark_png_path):
            raise Exception(translate('text', 'Failed to create image "{path}"').format(path=watermark_png_path))

        watermark_pdf_path = self.__main_window.cache.cache_path() + '/watermark.pdf'
        watermark_image = ImageReader(watermark_png_path)
        watermark_canvas = Canvas(watermark_pdf_path)
        document_page = self.__document.p_document.page(self.current_page_spinbox.value() - 1)
        document_page_size = document_page.pageSize()
        watermark_canvas.setPageSize((document_page_size.width(), document_page_size.height()))
        width = (self.__mark_widget.width() / self.view_label.width()) * document_page_size.width()
        height = (self.__mark_widget.height() / self.view_label.height()) * document_page_size.height()
        x = (self.__mark_widget.pos().x() / self.view_label.width()) * document_page_size.width()
        y = document_page_size.height() - (self.__mark_widget.pos().y() / self.view_label.height()) * \
            document_page_size.height() - height
        watermark_canvas.drawImage(image=watermark_image,
                                   x=x,
                                   y=y,
                                   width=width,
                                   height=height,
                                   mask='auto')
        watermark_canvas.save()

        base_document_name = QFileInfo(self.__main_window.document).baseName()
        document_dir_path = os.path.dirname(self.__main_window.document)
        preview_document_path = os.path.join(document_dir_path, base_document_name + '-preview.pdf')
        watermark_page_index = self.current_page_spinbox.value() - 1

        self.thread = WatermarkCreationThread(source_path=self.__document.path,
                                              watermark_path=watermark_pdf_path,
                                              preview_path=preview_document_path,
                                              watermark_page_index=watermark_page_index)
        self.thread.finished.connect(self.__on_watermark_creation_thread_finished)
        wait_dialog = WaitThreadDialog(thread=self.thread, parent=self)
        self.__main_window.start_waiting(wait_dialog)
        self.__block_control_widgets()
        self.thread.start()

    def __apply_stamp_push_button_clicked(self, _: bool) -> None:
        try:
            self.__create_preview_file()
        except Exception as exception:
            self.__main_window.show_error(translate('text', 'Merge watermark error'),
                                          str(exception))
            self.__main_window.stop_waiting()
            self.__block_control_widgets(False)

    def __on_current_page_spinbox_editing_finished(self) -> None:
        self.__load_page(self.current_page_spinbox.value())

    def __current_zoom_combobox_current_index_changed(self, index: int) -> None:
        self.__set_zoom_index(index)

    def __on_view_label_has_resize(self, old_size: QSize, size: QSize) -> None:
        self.__mark_widget.move(self.__mark_widget.pos().x() * (size.width() / old_size.width()),
                                self.__mark_widget.pos().y() * (size.height() / old_size.height()))
        self.__mark_widget.scale(self.current_zoom_combobox.currentData())

    def __on_main_window_selected_document_changed(self, _: str) -> None:
        if self.__main_window.tabWidget.currentWidget() != self:
            self.__unload_document()
        else:
            self.__find_and_load_document()

    def __combobox_certificates_current_index_changed(self, _: int) -> None:
        self.__unload_document()

    def __main_window_tab_widget_current_changed(self, index: int) -> None:
        if index != self.__main_window.tabWidget.indexOf(self):
            return

        if self.__document is not None:
            return

        self.__find_and_load_document_safe()

    def __block_control_widgets(self, block: bool = True) -> None:
        self.navigation_toolbar.setEnabled(not block)
        self.actions_toolbar.setEnabled(not block)

    def __load_document(self, path: str) -> None:
        self.__block_control_widgets()

        document = Poppler.Document.load(path)

        if document is None:
            raise Exception(translate('text', 'Couldn\'t open file "{path}" document').format(path=path))

        if document.isLocked():
            raise Exception(translate('text', 'Work with a read-encrypted file "{path}" is not supported')
                            .format(path=path))

        with open(path, 'rb') as file:
            pdf = PdfFileReader(file)
            if pdf.isEncrypted:
                raise Exception(translate('text', 'Work with encrypted for modification file "{path}" is not supported')
                                .format(path=path))

        document.setRenderHint(Poppler.Document.TextAntialiasing)
        self.__document = Document(document, path)
        self.current_page_spinbox.set_pages_count(document.numPages())

        self.__load_page(1)
        self.__load_mark(certificate=self.__main_window.selected_certificate,
                         template=self.selected_template)
        self.__mark_widget.move(0, 0)
        self.__block_control_widgets(False)

    def __load_mark(self, certificate: Optional[Certificate], template: Optional[TemplateData]) -> None:
        if self.__document is None:
            return

        if certificate is None:
            self.apply_stamp_push_button.setEnabled(False)
            self.__main_window.show_error(translate('text', 'Error'),
                                          translate('text', 'Certificate not selected'))
            return

        if template is None:
            self.apply_stamp_push_button.setEnabled(False)
            self.__main_window.show_error(translate('text', 'Error'),
                                          translate('text', 'Template not selected'))
            return

        self.__mark_widget.mark_label.template_text = template.read_html_text()
        self.__mark_widget.mark_label.style_sheet = template.read_css_text()
        self.__mark_widget.mark_label.certificate = certificate
        self.__mark_widget.scale(self.current_zoom_combobox.currentData())
        errors = self.__mark_widget.mark_label.errors
        if errors:
            keys_text = ', '.join(str(key_error.key) for key_error in errors)
            text = translate('text',
                             'Field(s) "{fields}" not filled in certificate "{certificate}". '
                             'Signature mark may not be displayed correctly. '
                             'Check config files "{path}"',
                             None,
                             len(errors)).format(fields=keys_text,
                                                 certificate=certificate,
                                                 path=template.path)
            self.__main_window.show_warning(translate('text', 'Warning'), text)
        self.apply_stamp_push_button.setEnabled(True)

    def __unload_document(self) -> None:
        self.__block_control_widgets()
        self.__document = None
        self.current_page_spinbox.set_pages_count(1)
        self.current_zoom_combobox.setCurrentIndex(0)
        self.view_label.clear()

    def __load_page(self, number: int) -> None:
        if self.__document is None:
            return

        index = number - 1
        if index < 0 or index > self.__document.p_document.numPages() - 1:
            return

        self.current_page_spinbox.setValue(number)
        dpi_x, dpy_y = self.__zoom_value_to_dpi(self.current_zoom_combobox.currentData())
        self.__load_page_on_view(index, dpi_x, dpy_y)

    def __set_zoom_index(self, index: int) -> None:
        if self.__document is None:
            return

        if index < 0 or index >= self.current_zoom_combobox.count():
            return

        if index != self.current_zoom_combobox.currentIndex():
            self.current_zoom_combobox.blockSignals(True)
            self.current_zoom_combobox.setCurrentIndex(index)
            self.current_zoom_combobox.blockSignals(False)

        page_index = self.current_page_spinbox.value() - 1
        dpi_x, dpy_y = self.__zoom_value_to_dpi(self.current_zoom_combobox.currentData())
        self.__load_page_on_view(page_index, dpi_x, dpy_y)

    def __load_page_on_view(self, index: int, dpi_x: int, dpi_y: int) -> None:
        page = self.__document.p_document.page(index)
        img = page.renderToImage(dpi_x, dpi_y)
        pixmap = QPixmap.fromImage(img)
        self.view_label.setPixmap(pixmap)

    def __current_screen(self) -> QScreen:
        return QApplication.screens()[QApplication.desktop().screenNumber(self)]

    def __zoom_value_to_dpi(self, value: float) -> Tuple[int, int]:
        current_screen = self.__current_screen()
        return current_screen.logicalDotsPerInchX() * value, current_screen.logicalDotsPerInchX() * value

    def __find_and_load_document(self) -> None:
        source_document_path = self.__main_window.document

        if source_document_path is None:
            self.__unload_document()
            return

        if not os.path.isfile(source_document_path):
            return

        base_document_name = QFileInfo(source_document_path).baseName()

        dir_path = os.path.dirname(source_document_path)
        local_preview_document_path = os.path.join(dir_path, base_document_name + '-preview.pdf')
        if os.path.isfile(local_preview_document_path):
            self.__load_document(local_preview_document_path)
            return

        cache_preview_document_path = os.path.join(self.__main_window.cache.cache_path(), base_document_name + '.pdf')
        if os.path.isfile(cache_preview_document_path):
            self.__load_document(cache_preview_document_path)
            return

        document_mimetype = Utils.file_mimetypes(source_document_path)
        if document_mimetype == 'application/pdf':
            self.__load_document(source_document_path)
            return

        if document_mimetype not in (self.__main_window.graphic_settings.pdf_conversion_mime_types or []):
            self.__main_window.show_error(translate('text', 'Can not load document'),
                                          translate('text', 'Document {path} not in pdf conversion mimetypes list')
                                          .format(path=source_document_path))
            return

        self.__pdf_conversion_process.setup_arguments(file_path=source_document_path,
                                                      outdir=self.__main_window.cache.cache_path())
        self.__main_window.start_waiting(InteractiveProcessDialog(process=self.__pdf_conversion_process, parent=self))
        self.__pdf_conversion_process.start()

    def __on_pdf_conversion_process_completed(self, result: CompletedResult) -> None:
        is_cancelled = self.__main_window.waiting_dialog.cancelled
        self.__main_window.stop_waiting()
        if is_cancelled:
            return

        if result.error is not None:
            self.__main_window.show_error(translate('text', 'Process error'),
                                          translate('text', 'Failed to create file'),
                                          self.__pdf_conversion_process.errorString())
            return
        if result.exit_status is not None and result.exit_status != QProcess.NormalExit:
            self.__main_window.show_error(translate('text', 'Process crashed'),
                                          translate('text', 'Failed to create file'),
                                          self.__pdf_conversion_process.errorString())
            return
        if result.exit_code is not None and result.exit_code != 0:
            self.__main_window.show_error(translate('text', 'Process exited with unexpected code: {code}')
                                          .format(code=result.exit_code),
                                          translate('text', 'Failed to create file'),
                                          Utils.to_utf8(self.__pdf_conversion_process.readAll()))
            return

        path = self.__pdf_conversion_process.output_file_path(file_path=self.__main_window.document,
                                                              outdir=self.__main_window.cache.cache_path())
        try:
            self.__load_document(path)
        except Exception as exception:
            self.__main_window.show_error(translate('text', 'Failed load document'),
                                          str(exception))

    def __on_watermark_creation_thread_finished(self) -> None:
        thread = self.sender()
        if type(thread) != WatermarkCreationThread:
            return

        self.__main_window.stop_waiting()

        if thread.error:
            self.__main_window.show_error(translate('text', 'Failed to create preview file "{path}"')
                                          .format(path=os.path.basename(thread.preview_path)),
                                          thread.error)
            self.__block_control_widgets(False)
            return

        if self.__main_window.container:
            container_creation_thread = ContainerCreationThread(container=self.__main_window.container,
                                                                files=[thread.preview_path],
                                                                update=True)
            container_creation_thread.finished.connect(self.__on_update_container_thread_finished)
            wait_dialog = WaitThreadDialog(thread=container_creation_thread, parent=self)
            self.__main_window.start_waiting(wait_dialog)
            container_creation_thread.start()
        else:
            self.__show_message(translate('text', 'Preview file created'),
                                translate('text', 'Preview file "{path}" created')
                                .format(path=thread.preview_path))
            old_page_index = self.current_page_spinbox.value()
            self.__unload_document()
            self.__find_and_load_document_safe()
            self.__load_page(old_page_index)

    def __find_and_load_document_safe(self) -> None:
        try:
            self.__find_and_load_document()
        except Exception as exception:
            self.__main_window.show_error(translate('text', 'Failed load document'),
                                          str(exception))

    def __show_message(self, title: str, text: str, details: str = None) -> None:
        msgbox = QMessageBox(self)
        msgbox.setWindowTitle(title)
        msgbox.setText(text)
        if details:
            msgbox.setDetailedText(details)
        msgbox.exec()

    def __on_update_container_thread_finished(self) -> None:
        thread = self.sender()
        if type(thread) != ContainerCreationThread:
            return

        self.__main_window.stop_waiting()
        if thread.error is not None:
            self.__main_window.show_error(translate('text', 'Container error'),
                                          translate('text', 'Failed to update container'),
                                          thread.error)
            return

        self.__show_message(translate('text', 'Preview file created'),
                            translate('text', 'Preview file created in "{container}" container')
                            .format(container=thread.container.path))
        old_page_index = self.current_page_spinbox.value()
        self.__unload_document()
        self.__find_and_load_document_safe()
        self.__load_page(old_page_index)
