# -*- coding: utf-8 -*-

import re
import os
import shutil
import yaml

from enum import IntEnum
from typing import List, Dict, Any, Optional

from PyQt5.QtCore import QCoreApplication

from libastra_csp.certificate import Certificate
from fly_csp_plugins.signature_mark_widget.certificate_info import CertificateInfo
from libastra_csp.utils import Utils


translate = QCoreApplication.translate


class CertificateKeyNotFoundError(Exception):
    def __init__(self, key: str, dictionary: Dict[str, Any]) -> None:
        self.__key = key
        self.__dictionary = dictionary

    @property
    def key(self) -> str:
        return self.__key

    @property
    def dictionary(self) -> Dict[str, Any]:
        return self.__dictionary


class TemplateData:
    __default_templates_dir_path = '/usr/share/astra-csp/sign-templates/'
    __global_templates_dir_path = '/etc/astra-csp/sign-templates/'
    __template_filename = 'template.html'
    __css_filename = 'template.css'
    __settings_filename = 'template.yaml'
    __name_key = 'name'

    class Location(IntEnum):
        Unknown = 0
        Default = 1
        Global = 2
        User = 3

    def __init__(self, name: str, location: Location) -> None:
        self.__name = name
        self.__location = location

    @classmethod
    def get_all_templates(cls) -> List[object]:
        templates = []
        for root, dirs, files in os.walk(TemplateData.default_dir_path()):
            for dir_name in dirs:
                templates.append(TemplateData(name=dir_name, location=TemplateData.Location.Default))
            break
        for root, dirs, files in os.walk(TemplateData.global_dir_path()):
            for dir_name in dirs:
                templates.append(TemplateData(name=dir_name, location=TemplateData.Location.Global))
            break
        for root, dirs, files in os.walk(TemplateData.user_dir_path()):
            for dir_name in dirs:
                templates.append(TemplateData(name=dir_name, location=TemplateData.Location.User))
            break

        return templates

    @classmethod
    def default_dir_path(cls) -> str:
        return cls.__default_templates_dir_path

    @classmethod
    def global_dir_path(cls) -> str:
        return cls.__global_templates_dir_path

    @staticmethod
    def user_dir_path() -> str:
        return os.path.join(os.path.expanduser('~'), '.local/share/rusbitech/fly-csp/sign-templates/')

    @property
    def dir_name(self) -> str:
        return self.__name

    @property
    def path(self) -> Optional[str]:
        if self.__location == TemplateData.Location.Default:
            return os.path.join(TemplateData.default_dir_path(), self.__name)
        elif self.__location == TemplateData.Location.Global:
            return os.path.join(TemplateData.global_dir_path(), self.__name)
        elif self.__location == TemplateData.Location.User:
            return os.path.join(TemplateData.user_dir_path(), self.__name)
        else:
            return None

    @property
    def exist(self) -> bool:
        return os.path.exists(self.path)

    @property
    def location(self) -> Location:
        return self.__location

    def read_html_text(self) -> str:
        return Utils.readall(os.path.join(self.path, self.__template_filename))

    def save_html_text(self, text: str) -> None:
        self.__create_dir()
        with open(os.path.join(self.path, self.__template_filename), 'w') as file:
            file.write(text)

    def read_css_text(self) -> str:
        return Utils.readall(os.path.join(self.path, self.__css_filename))

    def save_css_text(self, text: str) -> None:
        self.__create_dir()
        with open(os.path.join(self.path, self.__css_filename), 'w') as file:
            file.write(text)

    def read_name(self, lang: str = None) -> Optional[str]:
        with open(os.path.join(self.path, self.__settings_filename), 'r') as file:
            names = (yaml.load(file) or {}).get(self.__name_key)

        return names.get(lang, names.get(None)) if names else None

    def save_name(self, name: str, lang: str = None) -> None:
        self.__create_dir()
        path = os.path.join(self.path, self.__settings_filename)
        dictionary = {}
        try:
            with open(path, mode='r', encoding='utf-8') as file:
                dictionary.update(yaml.load(file) or {})
        except FileNotFoundError:
            pass
        with open(path, mode='w', encoding='utf-8') as file:
            dictionary.setdefault(self.__name_key, {})
            dictionary[self.__name_key][lang] = name
            yaml.dump(dictionary, stream=file)

    def delete(self) -> None:
        shutil.rmtree(self.path)

    def __create_dir(self) -> None:
        if not self.exist:
            os.makedirs(name=self.path, exist_ok=True)


class TemplateEngine:
    def __init__(self) -> None:
        self.__template = None
        self.__certificate = None
        self.__errors = []  # type: List[CertificateKeyNotFoundError]
        self.variable_re = re.compile(r'\$\w+(?:\$[\w_]+)*')

    @property
    def template(self) -> Optional[str]:
        return self.__template

    @template.setter
    def template(self, text: str) -> None:
        self.__template = text

    @property
    def certificate(self) -> Optional[Certificate]:
        return self.certificate

    @certificate.setter
    def certificate(self, certificate: Certificate) -> None:
        self.__certificate = certificate

    @property
    def errors(self) -> List[CertificateKeyNotFoundError]:
        return self.__errors

    def render(self) -> str:
        self.__errors.clear()

        cert = CertificateInfo(self.__certificate or {})
        template = self.__template or ''
        variables = self.variable_re.findall(template) or []
        for var in variables:
            try:
                value = self.__eval(var, cert)
            except CertificateKeyNotFoundError as exception:
                self.__errors.append(exception)
                value = ''
            template = template.replace(var, value)

        return template

    def __eval(self, variable: str, cert: CertificateInfo) -> str:
        nested_vars = [var.replace('_', ' ') for var in variable.split('$') if var]
        return self.__extract_nested_value(nested_vars, cert)

    def __extract_nested_value(self, keys: List[str], dictionary: Dict[str, Any]) -> str:
        key = keys.pop(0)

        if key not in dictionary:
            raise CertificateKeyNotFoundError(key=key, dictionary=dictionary)
        elif len(keys):
            return self.__extract_nested_value(keys, dictionary[key])
        else:
            return str(dictionary[key])
