# -*- coding: utf-8 -*-

import re

from typing import Tuple, Optional
from html.parser import HTMLParser


class TemplateHtmlParser(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self.__source_text = None  # type: Optional[str]
        self.__target_text = None  # type: Optional[str]
        self.__logo_path = None  # type: Optional[str]
        self.__logo_width = None  # type: Optional[int]
        self.__logo_height = None  # type: Optional[int]

    @property
    def source_text(self) -> Optional[str]:
        return self.__source_text

    @property
    def target_text(self) -> Optional[str]:
        return self.__target_text

    @property
    def logo_path(self) -> Optional[str]:
        return self.__logo_path

    @property
    def logo_width(self) -> Optional[int]:
        return self.__logo_width

    @property
    def logo_height(self) -> Optional[int]:
        return self.__logo_height

    @property
    def is_valid_template(self) -> bool:
        return self.__logo_path is not None and self.__logo_width is not None and self.__logo_height is not None

    def text_is_changed(self) -> bool:
        return self.__source_text != self.__target_text

    def set_logo_width(self, value: int) -> None:
        self.__logo_width = value
        self.__target_text = re.sub(r'img.+(width=".*?")',
                                    lambda match: match.group().replace(match.group(1), 'width="{value}"'
                                                                        .format(value=int(self.__logo_width))),
                                    self.__target_text)

    def set_logo_height(self, value: int) -> None:
        self.__logo_height = value
        self.__target_text = re.sub(r'img.+(height=".*?")',
                                    lambda match: match.group().replace(match.group(1), 'height="{value}"'
                                                                        .format(value=int(self.__logo_height))),
                                    self.__target_text)

    def set_logo_path(self, path: str) -> None:
        self.__logo_path = path
        self.__target_text = re.sub(r'img.+src=(".*?")',
                                    lambda match: match.group().replace(match.group(1), '"{path}"'
                                                                        .format(path=self.__logo_path)),
                                    self.__target_text)
        if self.logo_width is not None and self.logo_width == 0:
            self.set_logo_width(35)
        if self.logo_height is not None and self.logo_height == 0:
            self.set_logo_height(35)

    def remove_logo(self) -> None:
        self.__logo_path = ''
        self.__target_text = re.sub(r'img.+src=(".*?")',
                                    lambda match: match.group().replace(match.group(1), '""'),
                                    self.__target_text)
        self.set_logo_width(0)
        self.set_logo_height(0)

    def feed(self, data) -> None:
        self.__source_text = data
        self.__target_text = data
        super().feed(data)

    def handle_starttag(self, tag: str, attrs: Tuple[str, str]):
        if tag == 'img':
            images = [(first, second) for first, second in attrs if first == 'src']
            widths = [(first, second) for first, second in attrs if first == 'width']
            heights = [(first, second) for first, second in attrs if first == 'height']
            if images and len(images[0]) > 1:
                self.__logo_path = images[0][1]
            if widths and len(widths[0]) > 1:
                self.__logo_width = int(widths[0][1])
            if heights and len(heights[0]) > 1:
                self.__logo_height = int(heights[0][1])
