# -*- coding: utf-8 -*-

import os

from typing import Optional, List, Dict
from enum import IntEnum
from os.path import expanduser
from copy import deepcopy

from PyQt5 import uic
from PyQt5.QtCore import QProcess, Qt, QPoint, QUrl, QCoreApplication
from PyQt5.QtWidgets import QDialogButtonBox, QDialog, QComboBox, QPushButton, QLineEdit, QGroupBox, QApplication
from PyQt5.QtWidgets import QLabel, QScrollArea, QMessageBox, QAbstractButton, QTextEdit, QStackedWidget, QToolBar
from PyQt5.QtWidgets import QAction, QFileDialog, QSpinBox, QColorDialog, QMenu, QToolButton, QActionGroup
from PyQt5.QtGui import QIcon, QDesktopServices, QFont, QTextCharFormat, QTextCursor, QFontDatabase, QKeyEvent, QPixmap
from PyQt5.QtGui import QColor, QIntValidator

from libfly_csp.base_main_window import BaseMainWindow
from libastra_csp.certificate import Certificate
from libastra_csp.utils import Utils
from fly_csp_plugins.signature_mark_widget.mark import MarkLabel
from fly_csp_plugins.signature_mark_widget.template_engine import TemplateData
from fly_csp_plugins.signature_mark_widget.create_template_dialog import CreateTemplateDialog
from fly_csp_plugins.signature_mark_widget.template_html_parser import TemplateHtmlParser

translate = QCoreApplication.translate

dialog_ui = os.path.join(os.path.dirname(__file__), 'ui', 'templatessettingsdialog.ui')
UiFormClass, UiFormBase = uic.loadUiType(dialog_ui)  # type: QDialog, QDialog


class TemplatesSettingsDialog(UiFormClass, UiFormBase):
    class EditStackWidgetPage(IntEnum):
        Editor = 0
        NotValid = 1

    def __init__(self,
                 certificates: List[Certificate],
                 selected_certificate: Certificate,
                 main_window: BaseMainWindow) -> None:
        super().__init__(parent=main_window)
        self.template_combobox = None  # type: Optional[QComboBox]
        self.create_push_button = None  # type: Optional[QPushButton]
        self.delete_push_button = None  # type: Optional[QPushButton]
        self.name_label = None  # type: Optional[QLabel]
        self.name_line_edit = None  # type: Optional[QLineEdit]
        self.certificate_combobox = None  # type: Optional[QComboBox]
        self.preview_scroll_area = None  # type: Optional[QScrollArea]
        self.editor_group_box = None  # type: Optional[QGroupBox]
        self.stacked_widget = None  # type: Optional[QStackedWidget]
        self.warnings_label = None  # type: Optional[QLabel]
        self.text_edit = None  # type: Optional[QTextEdit]
        self.button_box = None  # type: Optional[QDialogButtonBox]
        self.setupUi(self)
        self.__tool_bar = QToolBar()
        self.__open_template_dir_action = None  # type: Optional[QAction]
        self.__width_image_spinbox = QSpinBox()
        self.__height_image_spinbox = QSpinBox()
        self.__insert_image_action = None  # type: Optional[QAction]
        self.__remove_image_action = None  # type: Optional[QAction]
        self.__bold_text_action = None  # type: Optional[QAction]
        self.__italic_text_action = None  # type: Optional[QAction]
        self.__underline_text_action = None  # type: Optional[QAction]
        self.__align_left_action = None  # type: Optional[QAction]
        self.__align_justify_action = None  # type: Optional[QAction]
        self.__align_center_action = None  # type: Optional[QAction]
        self.__align_right_action = None  # type: Optional[QAction]
        self.__color_action = None  # type: Optional[QAction]
        self.__insert_value_tool_button = QToolButton()
        self.__font_sizes_combobox = QComboBox()
        self.__preview_widget = MarkLabel()
        self.__certificates = certificates
        self.__main_selected_certificate = selected_certificate
        self.__main_window = main_window
        self.__selected_template_was_changed = False
        self.__skip_text_edit_set_text = False
        self.__change_certificate_in_progress = False
        self.__change_template_in_progress = False
        self.__setup()

    @property
    def selected_template(self) -> Optional[TemplateData]:
        return self.template_combobox.currentData()

    @property
    def html_text(self) -> str:
        return self.__preview_widget.template_text

    @property
    def css_text(self) -> str:
        return self.__preview_widget.style_sheet

    def set_selected_template(self, template: TemplateData) -> None:
        if not self.__selected_template_was_changed:
            indexes = [index for index in range(self.template_combobox.count())
                       if self.template_combobox.itemData(index).path == template.path]
            if indexes and indexes[0] >= 0:
                self.template_combobox.setCurrentIndex(indexes[0])

    def __setup(self) -> None:
        self.name_label.setText('{text}[{lang}]'.format(text=self.name_label.text(), lang=Utils.get_lang()))

        self.preview_scroll_area.setWidget(self.__preview_widget)

        self.__tool_bar.setWindowTitle(translate('text', 'Editor actions panel'))
        self.editor_group_box.layout().setMenuBar(self.__tool_bar)

        self.__open_template_dir_action = self.__tool_bar.addAction(QIcon.fromTheme('document-open-folder.png'),
                                                                    translate('text', 'Open template directory'))
        self.__open_template_dir_action.triggered.connect(self.__on_open_template_dir_action_triggered)

        self.__tool_bar.addSeparator()
        values_menu = QMenu()
        values_menu.aboutToShow.connect(self.__on_values_menu_about_to_show)
        self.__insert_value_tool_button.setText(translate('text', 'Inssert value'))
        self.__insert_value_tool_button.setToolTip(self.__insert_value_tool_button.text())
        self.__insert_value_tool_button.setIcon(QIcon.fromTheme('list-add'))
        self.__insert_value_tool_button.setMenu(values_menu)
        self.__insert_value_tool_button.setPopupMode(QToolButton.InstantPopup)
        insert_value_action = self.__tool_bar.addWidget(self.__insert_value_tool_button)
        insert_value_action.setText(self.__insert_value_tool_button.text())
        insert_value_action.setIcon(self.__insert_value_tool_button.icon())
        insert_value_action.setMenu(values_menu)

        self.__tool_bar.addSeparator()
        self.__insert_image_action = self.__tool_bar.addAction(QIcon.fromTheme('insert-image'),
                                                               translate('text', 'Insert logo'))
        self.__insert_image_action.triggered.connect(self.__on_insert_image_action_triggered)

        self.__remove_image_action = self.__tool_bar.addAction(QIcon.fromTheme('list-remove'),
                                                               translate('text', 'Remove logo'))
        self.__remove_image_action.triggered.connect(self.__on_remove_image_action_triggered)
        self.__width_image_spinbox.setMaximum(999)
        self.__width_image_spinbox.setToolTip(translate('text', 'Set logo width'))
        self.__width_image_spinbox.valueChanged.connect(self.__on_width_image_spinbox_value_changed)
        self.__tool_bar.addWidget(self.__width_image_spinbox)
        self.__height_image_spinbox.setMaximum(999)
        self.__height_image_spinbox.setToolTip(translate('text', 'Set logo height'))
        self.__height_image_spinbox.valueChanged.connect(self.__on_height_image_spinbox_value_changed)
        self.__tool_bar.addWidget(self.__height_image_spinbox)

        self.__tool_bar.addSeparator()
        self.__bold_text_action = self.__tool_bar.addAction(QIcon.fromTheme('format-text-bold'),
                                                            translate('text', 'Bold'))
        self.__bold_text_action.setCheckable(True)
        bold_font = self.__bold_text_action.font()
        bold_font.setBold(True)
        self.__bold_text_action.setFont(bold_font)
        self.__bold_text_action.triggered.connect(self.__on_bold_text_action_triggered)

        self.__italic_text_action = self.__tool_bar.addAction(QIcon.fromTheme('format-text-italic'),
                                                              translate('text', 'Italic'))
        self.__italic_text_action.setCheckable(True)
        italic_font = self.__italic_text_action.font()
        italic_font.setItalic(True)
        self.__italic_text_action.setFont(italic_font)
        self.__italic_text_action.triggered.connect(self.__on_italic_text_action_triggered)

        self.__underline_text_action = self.__tool_bar.addAction(QIcon.fromTheme('format-text-underline'),
                                                                 translate('text', 'Underline'))
        self.__underline_text_action.setCheckable(True)
        underline_font = self.__underline_text_action.font()
        underline_font.setUnderline(True)
        self.__underline_text_action.setFont(underline_font)
        self.__underline_text_action.triggered.connect(self.__on_underline_text_action_triggered)

        self.__align_left_action = self.__tool_bar.addAction(QIcon.fromTheme('format-justify-left'),
                                                             translate('text', 'Left'))
        self.__align_left_action.setCheckable(True)
        self.__align_center_action = self.__tool_bar.addAction(QIcon.fromTheme('format-justify-center'),
                                                               translate('text', 'Center'))
        self.__align_center_action.setCheckable(True)
        self.__align_right_action = self.__tool_bar.addAction(QIcon.fromTheme('format-justify-right'),
                                                              translate('text', 'Right'))
        self.__align_right_action.setCheckable(True)
        self.__align_justify_action = self.__tool_bar.addAction(QIcon.fromTheme('format-justify-fill'),
                                                                translate('text', 'Justify'))
        self.__align_justify_action.setCheckable(True)
        self.__align_justify_action.setVisible(False)
        align_group = QActionGroup(self)
        if QApplication.isLeftToRight():
            align_group.addAction(self.__align_left_action)
            align_group.addAction(self.__align_center_action)
            align_group.addAction(self.__align_right_action)
        else:
            align_group.addAction(self.__align_right_action)
            align_group.addAction(self.__align_center_action)
            align_group.addAction(self.__align_left_action)
        align_group.addAction(self.__align_justify_action)
        align_group.triggered.connect(self.__on_align_group_triggered)
        self.text_edit.cursorPositionChanged.connect(self.__on_text_edit_cursor_position_changed)

        self.__font_sizes_combobox.setEditable(True)
        validator = QIntValidator()
        validator.setBottom(0)
        validator.setTop(999)
        self.__font_sizes_combobox.setValidator(validator)
        for size in QFontDatabase.standardSizes():
            self.__font_sizes_combobox.addItem(str(size), float(size))
        self.__font_sizes_combobox.currentIndexChanged.connect(self.__on_font_sizes_combobox_current_index_changed)
        self.__tool_bar.addWidget(self.__font_sizes_combobox)

        pixmap = QPixmap(16, 16)
        pixmap.fill(Qt.black)
        self.__color_action = self.__tool_bar.addAction(QIcon(pixmap), translate('text', 'Color'))
        self.__color_action.triggered.connect(self.__on_color_action_triggered)

        self.text_edit.setContextMenuPolicy(Qt.CustomContextMenu)
        self.text_edit.customContextMenuRequested.connect(self.__on_text_edit_custom_context_menu_requested)
        self.text_edit.currentCharFormatChanged.connect(self.__on_text_edit_current_char_format_changed)

        self.template_combobox.currentIndexChanged.connect(self.__on_templates_combobox_current_index_changed)
        self.create_push_button.clicked.connect(self.__on_create_push_button_clicked)
        self.delete_push_button.clicked.connect(self.__on_delete_push_button_clicked)
        self.name_line_edit.textChanged.connect(self.__on_name_line_edit_text_changed)
        self.certificate_combobox.currentIndexChanged.connect(self.__on_certificate_combobox_current_index_changed)
        self.text_edit.textChanged.connect(self.__on_text_edit_text_changed)
        self.__preview_widget.content_changed.connect(self.__on_preview_widget_content_changed)
        self.button_box.clicked.connect(self.__on_button_box_clicked)

        for certificate in self.__certificates:
            self.certificate_combobox.addItem(str(certificate.subject), certificate)

        self.certificate_combobox.setCurrentIndex(self.certificate_combobox.findData(self.__main_selected_certificate))

        self.__load_templates()

    def reject(self) -> None:
        if self.__selected_template_was_changed:
            answer = QMessageBox.question(self,
                                          translate('text', 'Close editor'),
                                          translate('text', 'Are you sure you want to cancel changes and close form?'))
            if answer != QMessageBox.Yes:
                return

        super().reject()

    def accept(self) -> None:
        if not self.__selected_template_was_changed:
            super().accept()
        elif self.__save_changing():
            super().accept()

    # noinspection PyPep8Naming
    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
            return

        super().keyPressEvent(event)

    def __load_templates(self) -> None:
        self.template_combobox.clear()
        for template in TemplateData.get_all_templates():
            name = template.read_name(Utils.get_lang()) or template.dir_name
            if template.location == TemplateData.Location.User:
                text = translate('text', '{name} (current user template)').format(name=name)
            elif template.location == TemplateData.Location.Global:
                text = translate('text', '{name} (system template)').format(name=name)
            else:
                text = name
            self.template_combobox.addItem(text, template)

    @staticmethod
    def __extract_keys(value, key_string: str) -> Dict[str, str]:
        if issubclass(type(value), dict):
            results = {}
            for k, v in value.items():
                results.update(TemplatesSettingsDialog.__extract_keys(v, key_string + '$' + k))
            return results
        else:
            return {key_string.replace(' ', '_'): value}

    def __on_insert_image_action_triggered(self, _: bool) -> None:
        parser = TemplateHtmlParser()
        parser.feed(self.text_edit.toHtml())

        if not parser.is_valid_template:
            return

        if self.selected_template.exist:
            dir_path = self.selected_template.path
        else:
            dir_path = expanduser('~')

        dialog = QFileDialog(self, translate('text', 'Select image'), dir_path)
        extensions = ' '.join(['*.png'])
        dialog.setNameFilters([translate('text', 'Images ({extensions})').format(extensions=extensions)])
        if dialog.exec() != QDialog.Accepted:
            return

        parser.set_logo_path(dialog.selectedFiles()[0])

        if parser.text_is_changed():
            self.__preview_widget.template_text = parser.target_text
            self.__preview_widget.update_content()

    def __on_remove_image_action_triggered(self, _: bool) -> None:
        answer = QMessageBox.question(self,
                                      translate('text', 'Remove logo from template'),
                                      translate('text', 'Are you sure you want to remove image from template?'))
        if answer != QMessageBox.Yes:
            return

        parser = TemplateHtmlParser()
        parser.feed(self.text_edit.toHtml())

        if not parser.is_valid_template:
            return

        parser.remove_logo()

        if parser.text_is_changed():
            self.__preview_widget.template_text = parser.target_text
            self.__preview_widget.update_content()

    def __on_width_image_spinbox_value_changed(self, value: int) -> None:
        parser = TemplateHtmlParser()
        parser.feed(self.text_edit.toHtml())

        if not parser.is_valid_template:
            return

        parser.set_logo_width(value)

        self.__preview_widget.template_text = parser.target_text
        self.__preview_widget.update_content()

    def __on_height_image_spinbox_value_changed(self, value: int) -> None:
        parser = TemplateHtmlParser()
        parser.feed(self.text_edit.toHtml())

        if not parser.is_valid_template:
            return

        parser.set_logo_height(value)

        self.__preview_widget.template_text = parser.target_text
        self.__preview_widget.update_content()

    def __on_values_action_triggered(self, _: bool) -> None:
        action = self.sender()  # type: Optional[QAction]
        if not action:
            return

        text = action.data()  # type: Optional[str]
        if not text:
            return

        self.text_edit.insertPlainText(text)

    def __on_values_menu_about_to_show(self) -> None:
        menu = self.__insert_value_tool_button.menu()
        for action in menu.actions():
            action.deleteLater()
        menu.clear()

        certificate = self.certificate_combobox.currentData()  # type: Certificate

        for key, value in self.__extract_keys(certificate, '').items():
            action_text = menu.fontMetrics().elidedText('{value} - {key}'.format(value=value, key=key),
                                                        Qt.ElideMiddle,
                                                        500)
            action = QAction(action_text, menu)
            action.setData(key)
            action.triggered.connect(self.__on_values_action_triggered)
            menu.addAction(action)

    def __on_open_template_dir_action_triggered(self, _: bool) -> None:
        self.reject()
        QDesktopServices.openUrl(QUrl(self.selected_template.path))

    def __merge_current_char_format(self, char_format: QTextCharFormat) -> None:
        text_cursor = self.text_edit.textCursor()
        if not text_cursor.hasSelection():
            text_cursor.select(QTextCursor.WordUnderCursor)
        text_cursor.mergeCharFormat(char_format)
        self.text_edit.mergeCurrentCharFormat(char_format)

    def __on_bold_text_action_triggered(self, _: bool) -> None:
        char_format = self.text_edit.currentCharFormat()
        if self.__bold_text_action.isChecked():
            new_font_weight = QFont.Bold
        else:
            new_font_weight = QFont.Normal
        char_format.setFontWeight(new_font_weight)

        self.__merge_current_char_format(char_format)

    def __on_italic_text_action_triggered(self, _: bool) -> None:
        char_format = self.text_edit.currentCharFormat()
        char_format.setFontItalic(self.__italic_text_action.isChecked())

        self.__merge_current_char_format(char_format)

    def __on_underline_text_action_triggered(self, _: bool) -> None:
        char_format = self.text_edit.currentCharFormat()
        char_format.setFontUnderline(self.__underline_text_action.isChecked())

        self.__merge_current_char_format(char_format)

    def __on_align_group_triggered(self, action: QAction) -> None:
        if action == self.__align_left_action:
            self.text_edit.setAlignment(Qt.AlignLeft | Qt.AlignAbsolute)
        elif action == self.__align_center_action:
            self.text_edit.setAlignment(Qt.AlignHCenter)
        elif action == self.__align_right_action:
            self.text_edit.setAlignment(Qt.AlignRight | Qt.AlignAbsolute)
        elif action == self.__align_justify_action:
            self.text_edit.setAlignment(Qt.AlignJustify)

    def __text_edit_alignment_changed(self, alignment) -> None:
        if alignment & Qt.AlignLeft:
            self.__align_left_action.setChecked(True)
        elif alignment & Qt.AlignCenter:
            self.__align_center_action.setChecked(True)
        elif alignment & Qt.AlignRight:
            self.__align_right_action.setChecked(True)
        elif alignment & Qt.AlignJustify:
            self.__align_justify_action.setChecked(True)

    def __on_text_edit_cursor_position_changed(self) -> None:
        self.__text_edit_alignment_changed(self.text_edit.alignment())

    def __on_font_sizes_combobox_current_index_changed(self, index: int) -> None:
        if index < 0:
            return

        try:
            size = float(self.__font_sizes_combobox.itemData(index))
        except (ValueError, TypeError):
            try:
                size = float(self.__font_sizes_combobox.itemText(index))
            except (ValueError, TypeError):
                return

        if self.__font_sizes_combobox.itemData(index) != size:
            self.__font_sizes_combobox.setItemData(index, size)

        char_format = self.text_edit.currentCharFormat()
        if char_format.fontPointSize() != size:
            char_format.setFontPointSize(size)
            self.__merge_current_char_format(char_format)

    def __on_color_action_triggered(self, _: bool) -> None:
        color = QColorDialog.getColor(self.text_edit.textColor(), self)
        if not color.isValid():
            return

        char_format = self.text_edit.currentCharFormat()
        char_format.setForeground(color)
        self.__merge_current_char_format(char_format)

    def __on_text_edit_custom_context_menu_requested(self, position: QPoint) -> None:
        menu = QMenu()
        toolbar_actions = [action for action in self.__tool_bar.actions() if action.text() or action.isSeparator()]
        menu.addActions(toolbar_actions)
        menu.addSeparator()
        standard_menu = self.text_edit.createStandardContextMenu()
        menu.addActions(standard_menu.actions())

        menu.exec(self.text_edit.mapToGlobal(position))

    def __on_text_edit_current_char_format_changed(self, char_format: QTextCharFormat) -> None:
        self.__font_changed(char_format.font())
        self.__font_point_size_changed(char_format.fontPointSize())
        self.__color_changed(char_format.foreground().color())

    def __font_changed(self, font: QFont) -> None:
        self.__bold_text_action.setChecked(font.bold())
        self.__italic_text_action.setChecked(font.italic())
        self.__underline_text_action.setChecked(font.underline())

    def __font_point_size_changed(self, size: float) -> None:
        index = self.__font_sizes_combobox.findData(size)
        if index < 0:
            self.__font_sizes_combobox.addItem(str(size), size)
            self.__font_sizes_combobox.setCurrentIndex(self.__font_sizes_combobox.count() - 1)
        else:
            self.__font_sizes_combobox.setCurrentIndex(index)

    def __color_changed(self, color: QColor) -> None:
        pixmap = QPixmap(16, 16)
        pixmap.fill(color)
        self.__color_action.setIcon(QIcon(pixmap))

    def __on_templates_combobox_current_index_changed(self, _: int) -> None:
        if not self.selected_template:
            return

        self.__change_template_in_progress = True

        if self.selected_template.exist:
            self.__view_template(self.selected_template)

        self.__change_template_in_progress = False

    def __view_template(self, template: TemplateData) -> None:
        self.name_line_edit.setText(template.read_name(Utils.get_lang()))
        self.__update_preview_widget(template)

    def __update_preview_widget(self, template: TemplateData) -> None:
        try:
            self.__preview_widget.template_text = template.read_html_text()
            self.__preview_widget.style_sheet = template.read_css_text()
            self.__preview_widget.update_content()
        except Exception as exception:
            QMessageBox.critical(self,
                                 translate('text', 'Error'),
                                 translate('text', 'Can not update template preview. {text}')
                                 .format(text=str(exception)))

    def __on_create_push_button_clicked(self, _: bool) -> None:
        dialog = CreateTemplateDialog(parent=self)
        dialog.templates = [(self.template_combobox.itemText(index), self.template_combobox.itemData(index))
                            for index in range(self.template_combobox.count())]
        dialog.selected_template = self.selected_template

        if dialog.exec() != QDialog.Accepted:
            return

        self.template_combobox.addItem(translate('text', '{name} (not saved)').format(name=dialog.name),
                                       TemplateData(name=dialog.dir_name, location=dialog.selected_location))
        self.template_combobox.setCurrentIndex(self.template_combobox.count() - 1)
        self.name_line_edit.setText(dialog.name)

        self.__update_preview_widget(dialog.selected_template)

    def __on_delete_push_button_clicked(self, _: bool) -> None:
        answer = QMessageBox.question(self,
                                      translate('text', 'Delete template'),
                                      translate('text', 'Are you sure you want to delete template?'))
        if answer != QMessageBox.Yes:
            return

        index = self.template_combobox.currentIndex()

        if self.selected_template.exist:
            try:
                if self.selected_template.location == TemplateData.Location.User:
                    self.selected_template.delete()
                elif self.selected_template.location == TemplateData.Location.Global:
                    process = QProcess()
                    process.setProgram('pkexec')
                    arguments = ['/usr/libexec/flycsppreviewhelper',
                                 '--dir_name={dir_name}'.format(dir_name=self.selected_template.dir_name),
                                 '--delete']
                    process.setArguments(arguments)
                    process.start()
                    process.waitForFinished()
                    if process.exitStatus() != QProcess.NormalExit:
                        raise Exception(process.errorString())
                    elif process.exitCode() != 0:
                        output = Utils.to_utf8(process.readAllStandardError())
                        raise Exception(output if len(output) > 0 else process.errorString())
                else:
                    return
            except Exception as exception:
                QMessageBox.critical(self,
                                     translate('text', 'Remove template error'),
                                     translate('text', 'Can not remove template. {text}').format(text=str(exception)))
                return

        self.__selected_template_was_changed = False
        self.template_combobox.removeItem(index)

    def __on_name_line_edit_text_changed(self) -> None:
        if not self.__change_certificate_in_progress and not self.__change_template_in_progress:
            self.__selected_template_was_changed = True
        self.__update_dialog_state()

    def __on_certificate_combobox_current_index_changed(self, index: int) -> None:
        self.__change_certificate_in_progress = True
        self.__preview_widget.certificate = self.certificate_combobox.itemData(index)
        self.__preview_widget.update_content()
        self.__change_certificate_in_progress = False

    def __on_text_edit_text_changed(self) -> None:
        parser = TemplateHtmlParser()
        parser.feed(self.text_edit.toHtml())

        self.__insert_value_tool_button.setEnabled(parser.is_valid_template)
        self.__insert_image_action.setEnabled(parser.is_valid_template)
        self.__bold_text_action.setEnabled(parser.is_valid_template)
        self.__italic_text_action.setEnabled(parser.is_valid_template)
        self.__underline_text_action.setEnabled(parser.is_valid_template)
        self.__font_sizes_combobox.setEnabled(parser.is_valid_template)
        self.__color_action.setEnabled(parser.is_valid_template)
        self.__align_left_action.setEnabled(parser.is_valid_template)
        self.__align_center_action.setEnabled(parser.is_valid_template)
        self.__align_right_action.setEnabled(parser.is_valid_template)
        self.__align_justify_action.setEnabled(parser.is_valid_template)
        self.text_edit.setEnabled(parser.is_valid_template)

        has_image = parser.is_valid_template and bool(parser.logo_path)
        self.__remove_image_action.setEnabled(has_image)
        self.__width_image_spinbox.setEnabled(has_image)
        self.__height_image_spinbox.setEnabled(has_image)
        self.__width_image_spinbox.setValue(parser.logo_width if parser.is_valid_template else 0)
        self.__height_image_spinbox.setValue(parser.logo_height if parser.is_valid_template else 0)

        if not parser.is_valid_template:
            self.stacked_widget.setCurrentIndex(TemplatesSettingsDialog.EditStackWidgetPage.NotValid)
            return
        elif self.stacked_widget.currentIndex() != TemplatesSettingsDialog.EditStackWidgetPage.Editor:
            self.stacked_widget.setCurrentIndex(TemplatesSettingsDialog.EditStackWidgetPage.Editor)

        self.__skip_text_edit_set_text = True
        self.__preview_widget.template_text = parser.target_text
        self.__preview_widget.update_content()
        self.__update_dialog_state()

    def __on_preview_widget_content_changed(self) -> None:
        if not self.__change_certificate_in_progress and not self.__change_template_in_progress:
            self.__selected_template_was_changed = True

        if self.__preview_widget.errors:
            self.warnings_label.setText(translate('text',
                                                  'Field(s) "{fields}" not filled in certificate',
                                                  None,
                                                  len(self.__preview_widget.errors))
                                        .format(fields=', '.join(str(key_error.key) for key_error
                                                                 in self.__preview_widget.errors)))
        else:
            self.warnings_label.clear()

        if self.__skip_text_edit_set_text:
            self.__skip_text_edit_set_text = False
        else:
            self.text_edit.setText(self.__preview_widget.template_text)

    def __on_button_box_clicked(self, button: QAbstractButton) -> None:
        standard_button = self.button_box.standardButton(button)
        if standard_button == QDialogButtonBox.Cancel:
            self.reject()
        elif standard_button == QDialogButtonBox.Ok:
            self.accept()
        elif standard_button == QDialogButtonBox.Apply:
            if self.__save_changing():
                template = deepcopy(self.selected_template)
                self.__load_templates()
                self.set_selected_template(template)
        elif standard_button == QDialogButtonBox.RestoreDefaults:
            answer = QMessageBox.question(self,
                                          translate('text', 'Restore defaults'),
                                          translate('text', 'Are you sure you want to restore defaults content?'))
            if answer == QMessageBox.Yes:
                default_templates = [template for template in TemplateData.get_all_templates()
                                     if template.location == TemplateData.Location.Default]
                template = next(iter(default_templates))
                if template is None:
                    QMessageBox.critical(self,
                                         translate('text', 'Error'),
                                         translate('text', 'Default template not exists'))
                    return
                self.__update_preview_widget(template)

    def __save_changing(self) -> bool:
        try:
            if self.selected_template.location == TemplateData.Location.User:
                self.selected_template.save_html_text(self.html_text)
                self.selected_template.save_css_text(self.css_text)
                self.selected_template.save_name(self.name_line_edit.text(), Utils.get_lang())
            elif self.selected_template.location == TemplateData.Location.Global:
                process = QProcess()
                process.setProgram('pkexec')
                arguments = ['/usr/libexec/flycsppreviewhelper']
                arguments += ['--{key}={value}'.format(key=key, value=value)
                              for key, value in {
                                  'dir_name': self.selected_template.dir_name,
                                  'html': self.html_text,
                                  'css': self.css_text,
                                  'name': self.name_line_edit.text(),
                                  'name_lang':  Utils.get_lang()
                              }.items()]
                process.setArguments(arguments)
                process.start()
                process.waitForFinished()
                if process.exitStatus() != QProcess.NormalExit:
                    raise Exception(process.errorString())
                elif process.exitCode() != 0:
                    output = Utils.to_utf8(process.readAllStandardError())
                    raise Exception(output if len(output) > 0 else process.errorString())
            else:
                return False
        except Exception as exception:
            QMessageBox.critical(self,
                                 translate('text', 'Save template error'),
                                 translate('text', 'Can not save template. {text}').format(text=str(exception)))
            return False

        self.__selected_template_was_changed = False
        return True

    def __update_dialog_state(self) -> None:
        template_editable = self.selected_template and self.selected_template.location != TemplateData.Location.Default
        lock_another_template_editing = not self.__change_template_in_progress and self.__selected_template_was_changed
        template_exists = self.selected_template and self.selected_template.exist

        self.template_combobox.setEnabled(not lock_another_template_editing)
        self.create_push_button.setEnabled(not lock_another_template_editing)
        self.delete_push_button.setEnabled(template_editable and template_exists)

        template_index_valid = self.template_combobox.currentIndex() >= 0
        self.name_line_edit.setEnabled(template_index_valid and template_editable)
        self.certificate_combobox.setEnabled(template_index_valid)
        self.editor_group_box.setEnabled(template_index_valid and template_editable)

        self.__open_template_dir_action.setEnabled(template_exists)

        self.button_box.button(QDialogButtonBox.Ok).setEnabled(lock_another_template_editing)
        self.button_box.button(QDialogButtonBox.Apply).setEnabled(lock_another_template_editing)
        self.button_box.button(QDialogButtonBox.RestoreDefaults).setEnabled(template_editable)
